@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2019 gnome-mpv
+ * Copyright (c) 2014-2021 gnome-mpv
  *
  * This file is part of Celluloid.
  *
@@ -34,7 +34,10 @@ enum
 	PROP_SHOW_FULLSCREEN_BUTTON,
 	PROP_TIME_POSITION,
 	PROP_LOOP,
+	PROP_SHUFFLE,
 	PROP_VOLUME,
+	PROP_VOLUME_MAX,
+	PROP_VOLUME_POPUP_VISIBLE,
 	N_PROPERTIES
 };
 
@@ -48,6 +51,7 @@ struct _CelluloidControlBox
 	GtkWidget *next_button;
 	GtkWidget *previous_button;
 	GtkWidget *loop_button;
+	GtkWidget *shuffle_button;
 	GtkWidget *volume_button;
 	GtkWidget *fullscreen_button;
 	GtkWidget *seek_bar;
@@ -58,7 +62,10 @@ struct _CelluloidControlBox
 	gboolean show_fullscreen_button;
 	gdouble time_position;
 	gboolean loop;
+	gboolean shuffle;
 	gdouble volume;
+	gdouble volume_max;
+	gboolean volume_popup_visible;
 };
 
 struct _CelluloidControlBoxClass
@@ -163,10 +170,32 @@ set_property(	GObject *object,
 		self->loop = g_value_get_boolean(value);
 		break;
 
+		case PROP_SHUFFLE:
+		self->shuffle = g_value_get_boolean(value);
+		break;
+
 		case PROP_VOLUME:
 		self->volume = g_value_get_double(value);
 		break;
 
+		case PROP_VOLUME_MAX:
+		self->volume_max = g_value_get_double(value);
+		break;
+
+		case PROP_VOLUME_POPUP_VISIBLE:
+		self->volume_popup_visible = g_value_get_boolean(value);
+
+		if(	self->volume_popup_visible &&
+			gtk_widget_is_visible(GTK_WIDGET(self)) )
+		{
+			g_signal_emit_by_name(self->volume_button, "popup");
+		}
+		else if(!self->volume_popup_visible)
+		{
+			g_signal_emit_by_name(self->volume_button, "popdown");
+		}
+		break;
+
 		default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
 		break;
@@ -211,10 +240,22 @@ get_property(	GObject *object,
 		g_value_set_boolean(value, self->loop);
 		break;
 
+		case PROP_SHUFFLE:
+		g_value_set_boolean(value, self->shuffle);
+		break;
+
 		case PROP_VOLUME:
 		g_value_set_double(value, self->volume);
 		break;
 
+		case PROP_VOLUME_MAX:
+		g_value_set_double(value, self->volume_max);
+		break;
+
+		case PROP_VOLUME_POPUP_VISIBLE:
+		g_value_set_boolean(value, self->volume_popup_visible);
+		break;
+
 		default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
 		break;
@@ -419,6 +460,15 @@ celluloid_control_box_class_init(CelluloidControlBoxClass *klass)
 	g_object_class_install_property
 		(object_class, PROP_LOOP, pspec);
 
+	pspec = g_param_spec_boolean
+		(	"shuffle",
+			"Shuffle",
+			"Whether or not the the shuffle button is active",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property
+		(object_class, PROP_SHUFFLE, pspec);
+
 	pspec = g_param_spec_double
 		(	"volume",
 			"Volume",
@@ -430,6 +480,26 @@ celluloid_control_box_class_init(CelluloidControlBoxClass *klass)
 	g_object_class_install_property
 		(object_class, PROP_VOLUME, pspec);
 
+	pspec = g_param_spec_double
+		(	"volume-max",
+			"Volume Max",
+			"The maximum amplification level",
+			0.0,
+			G_MAXDOUBLE,
+			100.0,
+			G_PARAM_READWRITE );
+	g_object_class_install_property
+		(object_class, PROP_VOLUME_MAX, pspec);
+
+	pspec = g_param_spec_boolean
+		(	"volume-popup-visible",
+			"Volume popup visible",
+			"Whether or not the volume popup is visible",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property
+		(object_class, PROP_VOLUME_POPUP_VISIBLE, pspec);
+
 	g_signal_new(	"button-clicked",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
@@ -465,6 +535,9 @@ celluloid_control_box_class_init(CelluloidControlBoxClass *klass)
 static void
 celluloid_control_box_init(CelluloidControlBox *box)
 {
+	GtkWidget *popup = NULL;
+	GtkAdjustment *adjustment = NULL;
+
 	box->play_button = gtk_button_new();
 	box->stop_button = gtk_button_new();
 	box->forward_button = gtk_button_new();
@@ -473,6 +546,7 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	box->previous_button = gtk_button_new();
 	box->fullscreen_button = gtk_button_new();
 	box->loop_button = gtk_toggle_button_new();
+	box->shuffle_button = gtk_toggle_button_new();
 	box->volume_button = gtk_volume_button_new();
 	box->seek_bar = celluloid_seek_bar_new();
 	box->skip_enabled = FALSE;
@@ -482,7 +556,9 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	box->show_fullscreen_button = FALSE;
 	box->time_position = 0.0;
 	box->loop = FALSE;
+	box->shuffle = FALSE;
 	box->volume = 0.0;
+	box->volume_max = 100.0;
 
 	init_button(	box->play_button,
 			"media-playback-start-symbolic",
@@ -505,6 +581,9 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	init_button(	box->loop_button,
 			"media-playlist-repeat-symbolic",
 			_("Loop Playlist") );
+	init_button(	box->shuffle_button,
+			"media-playlist-shuffle-symbolic",
+			_("Shuffle Playlist") );
 	init_button(	box->fullscreen_button,
 			"view-fullscreen-symbolic",
 			_("Toggle Fullscreen") );
@@ -526,12 +605,24 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	gtk_container_add(GTK_CONTAINER(box), box->next_button);
 	gtk_box_pack_start(GTK_BOX(box), box->seek_bar, TRUE, TRUE, 0);
 	gtk_container_add(GTK_CONTAINER(box), box->loop_button);
+	gtk_container_add(GTK_CONTAINER(box), box->shuffle_button);
 	gtk_container_add(GTK_CONTAINER(box), box->volume_button);
 	gtk_container_add(GTK_CONTAINER(box), box->fullscreen_button);
 
+	popup =	gtk_scale_button_get_popup
+		(GTK_SCALE_BUTTON(box->volume_button));
+	adjustment =	gtk_scale_button_get_adjustment
+			(GTK_SCALE_BUTTON(box->volume_button));
+
+	g_object_bind_property(	popup, "visible",
+				box, "volume-popup-visible",
+				G_BINDING_DEFAULT );
 	g_object_bind_property(	box, "loop",
 				box->loop_button, "active",
 				G_BINDING_BIDIRECTIONAL );
+	g_object_bind_property(	box, "shuffle",
+				box->shuffle_button, "active",
+				G_BINDING_BIDIRECTIONAL );
 	g_object_bind_property_full(	box->volume_button, "value",
 					box, "volume",
 					G_BINDING_BIDIRECTIONAL,
@@ -539,6 +630,13 @@ celluloid_control_box_init(CelluloidControlBox *box)
 					mpv_to_gtk_volume,
 					NULL,
 					NULL );
+	g_object_bind_property_full(	box, "volume-max",
+					adjustment, "upper",
+					G_BINDING_DEFAULT|G_BINDING_SYNC_CREATE,
+					mpv_to_gtk_volume,
+					gtk_to_mpv_volume,
+					NULL,
+					NULL );
 
 	g_signal_connect(	box,
 				"button-press-event",
@@ -621,9 +719,7 @@ celluloid_control_box_get_volume(CelluloidControlBox *box)
 gboolean
 celluloid_control_box_get_volume_popup_visible(CelluloidControlBox *box)
 {
-	GtkScaleButton *volume_button = GTK_SCALE_BUTTON(box->volume_button);
-
-	return gtk_widget_is_visible(gtk_scale_button_get_popup(volume_button));
+	return box->volume_popup_visible;
 }
 
 void
@@ -640,7 +736,7 @@ celluloid_control_box_set_fullscreen_state(	CelluloidControlBox *box,
 					GTK_ICON_SIZE_BUTTON );
 
 	gtk_widget_set_visible(	box->fullscreen_button,
-				!fullscreen && box->show_fullscreen_button );
+				box->show_fullscreen_button );
 }
 
 void
